import base64
import uuid
from typing import Any, Dict, List, Optional

import requests

from PS3838.utils.constants import URL_PS3838


class Bet():
    """
    This class is used to place bets on the PS3838 API.
    """
    def __init__(
        self, 
        credentials: Dict[str, str],
        *args,
        **kwargs
    ):
        self.credentials = credentials
        self.base_url = URL_PS3838
        self.auth_header = self._get_auth_header()

    def _get_auth_header(self) -> Dict[str, str]:
        auth_str = f"{self.credentials['username']}:{self.credentials['password']}"
        b64_auth_str = base64.b64encode(auth_str.encode('utf-8')).decode('utf-8')
        return {"Authorization": f"Basic {b64_auth_str}", "Content-Type": "application/json"}
    
    def _make_request(
        self, 
        endpoint: str, 
        data: Dict[str, Any] = None, 
        params: Dict[str, Any] = None, 
        method: str = 'GET'
    ) -> Dict[str, Any]:
        url = f"{self.base_url}{endpoint}"
        
        try:
            if method == 'GET':
                response = requests.get(url, headers=self.auth_header, params=params)
            elif method == 'POST':
                response = requests.post(url, headers=self.auth_header, json=data)
            response.raise_for_status()  # Raise an exception for HTTP errors
            return response.json()  # Try to parse JSON response
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.RequestException as req_err:
            print(f"Request error occurred: {req_err}")
        except requests.exceptions.JSONDecodeError as json_err:
            print(f"JSON decode error occurred: {json_err}")
            print(f"Response content: {response.text}")
        
        return {}

    def get_bets_v3(
        self, 
        betlist: Optional[str] = None, 
        from_date: Optional[str] = None, 
        to_date: Optional[str] = None,
        bet_ids: List[int] = None, 
        unique_request_ids: List[str] = None
    ) -> Dict[str, Any]:
        '''
        Get a list of bets that are already placed

        parameters:
            betlist: 'ALL' or 'SETTLED' or 'PENDING' or 'CANCELLED' or 'ALL_PENDING' (Optional)
            from_date: ISO8601 format (REQUIRED when betlist is provided)
            to_date: ISO8601 format (REQUIRED when betlist is provided)
            bet_ids: list of bet IDs
            unique_request_ids: list of unique request IDs
        
        return:
            JSON response
        '''

        params = {
            'betlist': betlist,
            'fromDate': from_date,
            'toDate': to_date,
            'betIds': ','.join(map(str, bet_ids)) if bet_ids else None,
            'uniqueRequestIds': ','.join(unique_request_ids) if unique_request_ids else None
        }
        return self._make_request('/v3/bets', params=params)

    def place_bet(
        self, 
        stake: float, 
        line_id: int, 
        sport_id: int, 
        event_id: int, 
        period_number: int, 
        team: str, 
        side: str = None, 
        handicap: float = None
    ) -> Dict[str, Any]:
        """
        Place a bet on the PS3838 API.
        
        Parameters:
            stake: The amount of the bet (min 5 euros)
            line_id: The line ID of the bet
            sport_id: The sport ID of the event (football = 29)
            event_id: The event ID of the event
            period_number: The period number of the event (0 for full time)
            team: The team to bet on "TEAM1" or "TEAM2" or "DRAW"
            side: The side to bet on (optional)
            handicap: The handicap value (optional)
            Other values that will be passed to the API are hardcoded
            
        Returns:
            The response from the API
        """

        # Generate a unique request ID
        unique_request_id = str(uuid.uuid4())  
        
        # Construct the data payload
        data = {
            "oddsFormat": "DECIMAL",
            "uniqueRequestId": unique_request_id,
            "acceptBetterLine": True,
            "stake": stake,
            "winRiskStake": "RISK",
            "lineId": line_id,
            "altLineId": None,
            "pitcher1MustStart": True,
            "pitcher2MustStart": True,
            "fillType": "NORMAL",
            "sportId": sport_id,
            "eventId": event_id,
            "periodNumber": period_number,
            "betType": "MONEYLINE",
            "team": team,
            "side": side,
            "handicap": handicap
        }

        return self._make_request('/v2/bets/place', data=data, method='POST')

    def check_maintenance(self) -> Dict[str, Any]:
        """
        Check the maintenance status of the PS3838 API. If the status is 'ALL_BETTING_ENABLED', then betting is allowed.
        Else, it means that the API is in maintenance mode and betting is disabled.
        """
        return self._make_request('/v1/bets/betting-status')



import base64
from typing import Any, Dict, List, Optional

import requests

from PS3838.utils.constants import URL_PS3838


class Retrieve():
    def __init__(
        self,
        credentials : Dict[str, str], 
        *args,
        **kwargs,
    ):
        self._credentials = credentials
        self._base_url = URL_PS3838
        self._auth_header = self._get_auth_header()

    def _get_auth_header(self) -> Dict[str, str]:
        auth_str = f"{self._credentials['username']}:{self._credentials['password']}"
        b64_auth_str = base64.b64encode(auth_str.encode('utf-8')).decode('utf-8')
        return {"Authorization": f"Basic {b64_auth_str}"}

    def _make_request(
        self, 
        endpoint: str, 
        params: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        url = f"{self._base_url}{endpoint}"
        response = requests.get(url, headers=self._auth_header, params=params)
        
        try:
            response.raise_for_status()  # Raise an exception for HTTP errors
            return response.json()  # Try to parse JSON response
        except requests.exceptions.HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
        except requests.exceptions.RequestException as req_err:
            print(f"Request error occurred: {req_err}")
        except requests.exceptions.JSONDecodeError as json_err:
            print(f"JSON decode error occurred: {json_err}")
            print(f"Response content: {response.text}")
        
        return {}

    def get_fixtures_v3(
        self, 
        sport_id: int, 
        league_ids: Optional[List[int]] = None,
        is_live: Optional[bool] = None, 
        since: Optional[int] = None,
        event_ids: Optional[List[int]] = None
    ) -> Dict[str, Any]:
        params = {
            "sportId": sport_id,
            "leagueIds": ",".join(map(str, league_ids)) if league_ids else None,
            "isLive": 1 if is_live else None,
            "since": since,
            "eventIds": ",".join(map(str, event_ids)) if event_ids else None
        }
        return self._make_request("/v3/fixtures", params)

    def get_straight_line_v2(
        self, 
        league_id : int,
        handicap: float = None,
        odds_format: str = "American",
        sport_id: int = None,
        event_id: int = None,
        period_number: int = None,
        bet_type: str = None,
        team : str = None,
        side: Optional[str] = None,
    ) -> Dict[str, Any]:
        params = {
            "leagueId": league_id,
            "handicap": handicap,
            "oddsFormat": odds_format,
            "sportId": sport_id,
            "eventId": event_id,
            "periodNumber": period_number,
            "betType": bet_type,
            "team": team,
            "side": side
        }
        return self._make_request("/v2/line", params)

    def get_sports_v3(self) -> Dict[str, Any]:
        return self._make_request("/v3/sports")

    def get_leagues_v3(self, sport_id: int) -> Dict[str, Any]:
        params = {"sportId": sport_id}
        return self._make_request("/v3/leagues", params)




from PS3838.utils.tools_code import find_best_match


def test_find_best_match_exact_match():
    assert find_best_match("Argentina", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_case_insensitive_match():
    assert find_best_match("argentina", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_with_spaces():
    assert find_best_match("Argen tina", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_with_punctuation():
    assert find_best_match("Argentina!", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_partial_match():
    assert find_best_match("Argen", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_no_close_match():
    assert find_best_match("Chile", ["Argentina", "Brazil", "Colombia"]) is None

def test_find_best_match_empty_candidates():
    assert find_best_match("Argentina", []) is None

def test_find_best_match_empty_name():
    assert find_best_match("", ["Argentina", "Brazil", "Colombia"]) is None

def test_find_best_match_multiple_close_matches():
    assert find_best_match("Argentin", ["Argentina", "Argentinos", "Brazil"]) == "Argentina"

def test_find_best_match_with_numbers():
    assert find_best_match("Argentina123", ["Argentina", "Brazil123", "Colombia"]) == "Argentina"

def test_find_best_match_all_special_characters():
    assert find_best_match("@rgentina!", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_cutoff_no_match():
    assert find_best_match("Arg", ["Argentina", "Brazil", "Colombia"]) == "Argentina"

def test_find_best_match_levenshtein():
    assert find_best_match("Arg", ["Argentina", "Brazil", "Colombia"]) == "Argentina"
    assert find_best_match("Brazil", ["Argentina", "Brazil", "Colombia"]) == "Brazil"
    assert find_best_match("Chil", ["Argentina", "Brazil", "Colombia", "Chile"]) == "Chile"
    assert find_best_match("Chile", ["Argentina", "Brazil", "Colombia"]) is None


import base64
import uuid
from unittest.mock import Mock, patch

import pytest

from PS3838.PS3838Bet import Bet
from PS3838.utils.constants import URL_PS3838


@pytest.fixture
def credentials():
    return {"username": "test_user", "password": "test_pass"}

@pytest.fixture
def bet_instance(credentials):
    return Bet(credentials=credentials)

def test_get_auth_header(bet_instance, credentials):
    expected_auth_str = f"{credentials['username']}:{credentials['password']}"
    expected_b64_auth_str = base64.b64encode(expected_auth_str.encode('utf-8')).decode('utf-8')
    expected_auth_header = {"Authorization": f"Basic {expected_b64_auth_str}", "Content-Type": "application/json"}

    assert bet_instance.auth_header == expected_auth_header

@patch('requests.get')
def test_make_request_get(mock_get, bet_instance):
    mock_response = Mock()
    expected_json = {"key": "value"}
    mock_response.json.return_value = expected_json
    mock_response.raise_for_status = Mock()
    mock_get.return_value = mock_response

    endpoint = "/test"
    params = {"param": "value"}
    response = bet_instance._make_request(endpoint, params=params, method='GET')

    mock_get.assert_called_once_with(f"{URL_PS3838}{endpoint}", headers=bet_instance.auth_header, params=params)
    assert response == expected_json

@patch('requests.post')
def test_make_request_post(mock_post, bet_instance):
    mock_response = Mock()
    expected_json = {"key": "value"}
    mock_response.json.return_value = expected_json
    mock_response.raise_for_status = Mock()
    mock_post.return_value = mock_response

    endpoint = "/test"
    data = {"data": "value"}
    response = bet_instance._make_request(endpoint, data=data, method='POST')

    mock_post.assert_called_once_with(f"{URL_PS3838}{endpoint}", headers=bet_instance.auth_header, json=data)
    assert response == expected_json

@patch('PS3838.PS3838Bet.Bet._make_request')
def test_get_bets_v3(mock_make_request, bet_instance):
    expected_response = {"bets": []}
    mock_make_request.return_value = expected_response

    betlist = "test_betlist"
    from_date = "2024-01-01"
    to_date = "2024-01-02"
    bet_ids = [1, 2, 3]
    unique_request_ids = ["id1", "id2"]
    response = bet_instance.get_bets_v3(betlist, from_date, to_date, bet_ids, unique_request_ids)

    mock_make_request.assert_called_once_with(
        '/v3/bets',
        params={
            'betlist': betlist,
            'fromDate': from_date,
            'toDate': to_date,
            'betIds': '1,2,3',
            'uniqueRequestIds': 'id1,id2'
        }
    )
    assert response == expected_response

@patch('PS3838.PS3838Bet.Bet._make_request')
def test_place_bet(mock_make_request, bet_instance):
    expected_response = {
        "status": "ACCEPTED",
        "straightBet": {
            "betId": 759629245,
            "wagerNumber": 1,
            "betStatus": "ACCEPTED"
        }
    }
    mock_make_request.return_value = expected_response

    odds_format = "Decimal"
    stake = 5.0
    line_id = 12345
    sport_id = 29
    event_id = 98765
    period_number = 0
    bet_type = "MONEYLINE"
    team = "Team1"
    response = bet_instance.place_bet(odds_format, stake, line_id, sport_id, event_id, period_number, bet_type, team)

    data = {
        "oddsFormat": odds_format,
        "uniqueRequestId": str(uuid.uuid4()),  # This needs to be dynamically generated, so we can't assert it directly
        "acceptBetterLine": True,
        "stake": stake,
        "winRiskStake": "RISK",
        "lineId": line_id,
        "altLineId": None,
        "pitcher1MustStart": True,
        "pitcher2MustStart": True,
        "fillType": "NORMAL",
        "sportId": sport_id,
        "eventId": event_id,
        "periodNumber": period_number,
        "betType": bet_type,
        "team": team,
        "side": None,
        "handicap": None
    }

    mock_make_request.assert_called_once()
    called_args, called_kwargs = mock_make_request.call_args
    assert called_args == ('/v2/bets/place',)
    assert called_kwargs['data'].keys() == data.keys()  # We check the keys instead of values because uniqueRequestId is dynamic
    assert called_kwargs['method'] == 'POST'

    assert response == expected_response

@patch('PS3838.PS3838Bet.Bet._make_request')
def test_check_maintenance(mock_make_request, bet_instance):
    expected_response = {"status": "ALL_BETTING_ENABLED"}
    mock_make_request.return_value = expected_response

    response = bet_instance.check_maintenance()

    mock_make_request.assert_called_once_with('/v1/bets/betting-status')
    assert response == expected_response


from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Tuple

import Levenshtein

from PS3838.PS3838Bet import Bet
from PS3838.PS3838Retrieve import Retrieve
from PS3838.telegram.telegram_bot import CustomLogger
from PS3838.utils.constants import ID_SOCCER
from PS3838.utils.errors import RetrieveMatchError


def retrieve_matches(
    list_matches : List[Dict[str, Any]],
    api_retrieve : Retrieve,
    logger : Optional[CustomLogger] = None,
    *args, 
    **kwargs
) -> List[Dict[str, Any]]:
    """
    This function retrieves the matches from the list of matches provided. It retrieves the matches one by one and appends them to a list of matches found.
    
    Parameters:
        list_matches : List[Dict[str, Any]] : List of matches to retrieve.
        api_retrieve : Retrieve : API object to retrieve the matches
        logger : Optional[CustomLogger] : Logger to log the results
    
    Returns:
        matches_found : List[Dict[str, Any]] : List of matches found
    """

    # Retrieve the matches
    matches_found = []
    for match in list_matches:
        
        try : 
            # Try to find the match that corresponds to the parameters
            match_found= retrieve_match(api_retrieve, league=match["league"], name_team1=match["team1"], name_team2=match["team2"], date=match["date"] if "date" in match.keys() else None, result=match["result"] if "result" in match.keys() else None, amount=match["amount"] if "amount" in match.keys() else None, odd_min=match["odd_min"] if "odd_min" in match.keys() else None, logger=logger)


            # Append the match to the list of matches found
            matches_found.append(match_found)
            if logger:
                logger.info(f"Match found for {match_found['home']} - {match_found['away']}")

        except Exception as e:
            if logger:
                logger.error(f"Error retrieving match for {match['team1']} - {match['team2']} : {e}")
            continue

    return matches_found

def retrieve_match(
    api_retrieve : Retrieve,
    league : str | int, 
    name_team1 : str, 
    name_team2 : str, 
    date : Optional[datetime] = None, 
    result : Optional[int] = None,
    amount : Optional[int] = None,
    odd_min : Optional[float] = None,
    logger : Optional[CustomLogger] = None,
    *args, 
    **kwargs
) -> Dict[str, Any]:
    """
    This function retrieves a match from the API based on the parameters provided. It retrieves the leagues, the matches in the league, and then finds the match that corresponds to the parameters.
    
    Parameters:
        api_retrieve : Retrieve : API object to retrieve the matches
        league : str | int : League name or ID (better to put the ID)
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        date : Optional[datetime] : Date of the match
        result : Optional[int] : Result of the match (1 = team1 win, 2 = team2 win, 0 = draw)
        amount : Optional[int] : Amount to bet on the match
        odd_min : Optional[float] : Minimum odds to bet on the match
        logger : Optional[CustomLogger] : Logger to log the results
        
    Returns:
        match : Dict[str, Any] : Match found
    """

    # Find the leagues
    leagues = api_retrieve.get_leagues_v3(sport_id=ID_SOCCER)

    # Find the ID of the league
    league_id = find_league_id(leagues, league, logger)

    # Find the matches in the league
    fixtures = api_retrieve.get_fixtures_v3(sport_id=ID_SOCCER, league_ids=[league_id])

    # Keep only the elements where 'statut' = 'O' which means the match is open
    fixtures = [match for match in fixtures["league"][0]["events"] if match["status"] == "O"]

    print(fixtures)

    # Find the match
    match = find_match(fixtures, name_team1, name_team2, date, logger)

    # Verify if the parameter result is correct
    result = verify_result(match, name_team1, name_team2, result, logger)

    # Add the league ID to the match, the result, the amount and the minimum odds that are fundamentals parameters if we want to bet afterwards
    match["league"] = league_id
    match["result"] = result
    match["amount"] = amount
    match["odd_min"] = odd_min

    return match

def get_team_odds(api_retrieve: Retrieve, match: Dict[str, Any], team: str) -> Dict[str, Any]:
    """
    This function retrieves the odds for a specific team in a match.

    Parameters:
        api_retrieve : Retrieve : API object to retrieve the odds
        match : Dict[str, Any] : Match to retrieve the odds for
        team : str : Team to retrieve the odds for

    Returns:
        odds : Dict[str, Any] : Odds for the team
    """
    odds_response = api_retrieve.get_straight_line_v2(
        league_id=match["league"],
        handicap=0,
        odds_format="Decimal",
        sport_id=ID_SOCCER,
        event_id=match["id"],
        period_number=match.get("period_number", 0),
        bet_type="MONEYLINE",
        team=team
    )

    return odds_response.get("price", 0), odds_response.get("lineId", 0)
    
def normalize_name(name : str) -> str:
    """
    This function normalizes a name by removing all special characters and converting it to lowercase.
    """
    return ''.join(e for e in name.lower() if e.isalnum())


def find_best_match(name: str, candidates: list[str], cutoff : float = 0.5) -> str:
    """
    This function finds the best match for a name in a list of candidates. It uses the Levenshtein distance to calculate the similarity between the name and the candidates. It is useful when the name of sport_api is not exactly the same as the name on PS3838. The cutoff parameter is used to filter the candidates that are too different from the name.
    """
    normalized_name = normalize_name(name)
    normalized_candidates = [normalize_name(candidate) for candidate in candidates]

    best_match = None
    highest_similarity = -1

    for candidate in normalized_candidates:
        similarity = Levenshtein.ratio(normalized_name, candidate)
        if similarity > highest_similarity and similarity >= cutoff:
            highest_similarity = similarity
            best_match = candidate

    if best_match:
        best_match_index = normalized_candidates.index(best_match)
        return candidates[best_match_index]

    return None

def find_match(
    fixtures : list[Dict[str, Any]], 
    name_team1 : str, 
    name_team2 : str, 
    date : datetime = None, 
    logger : CustomLogger = None
) -> Dict[str, Any]:
    """
    This function finds the match that corresponds to the team names and the date provided. It first tries to have the perfect match between team1 and team2 and the date, and then tries a lighter version without the date.
    
    Parameters:
        fixtures : list[Dict[str, Any]] : List of available matches returned by PS3838.
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        date : datetime : Date of the match
        logger : CustomLogger : Logger to log the results
        
    Returns:
        match : Dict[str, Any] : Match found
    """
    # Find the candidates for the team names
    candidates_home = []
    candidates_away = []
    for match in fixtures:
        candidates_home.append(match["home"])
        candidates_away.append(match["away"])

    match = None
    # Find the match
    for fixture in fixtures:
        if (find_best_match(name_team1, candidates_home) == fixture["home"] and find_best_match(name_team2, candidates_away) == fixture["away"]) or (find_best_match(name_team1, candidates_away) == fixture["away"] and find_best_match(name_team2, candidates_home) == fixture["home"]) and datetime.fromisoformat(fixture["starts"].replace('Z','')) == date:
            match = fixture
            break
        elif (find_best_match(name_team1, candidates_home) == fixture["home"] and find_best_match(name_team2, candidates_away) == fixture["away"]) or (find_best_match(name_team1, candidates_away) == fixture["away"] and find_best_match(name_team2, candidates_home) == fixture["home"]):
            match = fixture
            break
        
    if not match:
        if logger:
            logger.error(f"Match not found for {name_team1} - {name_team2}")
        raise RetrieveMatchError("Match not found")
    
    return match

def verify_result(
    match : Dict[str, Any],
    name_team1 : str, 
    name_team2 : str, 
    result : int, 
    logger : CustomLogger = None
) -> int:
    """
    This function verifies if the result corresponds to the home team or the away team, and changes it if necessary.
    For example if we entered Marseille - Lens with result 1 and the actual match is Lens - Marseille, the function will return 2.
    
    Parameters:
        match : Dict[str, Any] : Match found
        name_team1 : str : Name of the first team
        name_team2 : str : Name of the second team
        result : int : Result of the match (1 = team1 win, 2 = team2 win, 0 = draw)
        logger : CustomLogger : Logger to log the results
        
    Returns:
        int : 1 if the result corresponds to the home team, 2 if the result corresponds to the away team, 0 if the result corresponds to a draw
    """
    if result is not None:
        if result == 1:
            if (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 2
            else:
                if logger:
                    logger.error(f"Error while verifying if the result 1 corresponds to the home team")
                raise RetrieveMatchError("Error while verifying if the result 1 corresponds to the home team")
        elif result == 2:
            if (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) and (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 1
            elif (find_best_match(match["home"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["away"], [name_team1, name_team2]) == name_team2):
                return 2
            elif (find_best_match(match["away"], [name_team1, name_team2]) == name_team1) or (find_best_match(match["home"], [name_team1, name_team2]) == name_team2):
                return 1
            else:
                if logger:
                    logger.error(f"Error while verifying if the result 2 corresponds to the home team")
                raise RetrieveMatchError("Error while verifying if the result 2 corresponds to the home team")
        else:
            return 0
    else:
        return None
    

# To be corrected in the future
def find_league_id(
    leagues: Dict[str, Any], 
    league: str | int, 
    logger : CustomLogger = None
) -> int:
    """
    This function finds the ID of the league based on the name or the ID provided.
    The possibility of putting a string will be removed in the future to use directly the ID of the league."""
    if isinstance(league, int):
        # If league is an integer, directly find the matching ID
        for element in leagues.get("leagues", []):
            if element["id"] == league:
                return element["id"]
    else:
        # If league is a string, find the best matching league name
        league_names = [element["name"] for element in leagues.get("leagues", [])]
        league_name = find_best_match(league, league_names)
        for element in leagues.get("leagues", []):
            if element["name"] == league_name:
                return element["id"]
    if logger:
        logger.error(f"League not found")
    raise RetrieveMatchError("League not found")

def change_date(date : str, days : int, hours : int) -> str:
    """
    This function changes the date by adding days and hours to it. It is useful to search for bets that were placed from 25 days in the past to 1 day in the future (as it is done in the function is_already_bet).
    """
    date = datetime.fromisoformat(date.replace('Z',''))
    date = date + timedelta(days=days, hours=hours)
    return date.isoformat() + 'Z'

def place_bets(
    match_odds : List[Tuple[Dict[str, Any], Dict[str, Any]]], 
    bet : Bet, 
    logger : CustomLogger = None
):
    """
    This function places the bets for the matches found with the odds provided. It verifies if the odds are not too low and if the bet is not already placed. Then in function of the parameter result, it places the bet on the home team, the away team or the draw.
    
    Parameters:
        match_odds : List[Tuple[Dict[str, Any], Dict[str, Any]]] : List of matches found with the odds
        bet : Bet : API object to place the bets
        logger : CustomLogger : Logger to log the results
    """
    for match, odds in match_odds:
        if is_already_bet(bet, match):
            if logger:
                logger.warning(f"Bet already placed for match {match['id']} : {match['home']} - {match['away']}, skipping bet")
            continue
        if match["result"] == 1:
            if not verify_odds(match["odd_min"], odds["team1_odds"]):
                if logger:
                    logger.warning(f"Odds for match {match['id']} are finally too low, skipping bet")
                continue
            bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="TEAM1")
        elif match["result"] == 2:
            if not verify_odds(match["odd_min"], odds["team2_odds"]):
                if logger:
                    logger.warning(f"Odds for match {match['id']} are finally too low, skipping bet")
                continue
            bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="TEAM2")
        elif match["result"] == 0:
            if not verify_odds(match["odd_min"], odds["draw_odds"]):
                if logger:
                    logger.warning(f"Odds for match {match['id']} are finally too low, skipping bet")
                continue
            bet.place_bet(stake=match["amount"], line_id=match["line_id"], sport_id=ID_SOCCER, event_id=match["id"], period_number=0, team="Draw")
        if is_already_bet(bet, match):
            if logger:
                logger.info(f"Bet placed successfully for match {match['home']} - {match['away']}")

def verify_odds(odd_min : float, real_odd : float) -> bool:
    """
    This function verifies if the real odd is higher than the minimum odd required to place the bet. It is useful to check this right before placing the bet, to avoid placing a bet with too low odds.
    """
    return real_odd >= odd_min

def is_already_bet(api_bet : Bet, match : Dict[str, Any]) -> bool:
    """
    This function verifies if the bet is already placed for the match provided. It retrieves the bets placed in the past and checks if the bet is here. Also depending on the status of the bet, it returns True or False.
    """
    bets = api_bet.get_bets_v3(betlist='ALL', from_date=change_date(date=match["starts"], days=-25, hours=0), to_date=change_date(date=match["starts"], days=1, hours=0))
    if "straightBets" in bets.keys():
        straights = bets["straightBets"]
        for bet in straights:
            if (bet["eventId"] == match["id"]) and (bet['betStatus'] == 'LOSE' or bet['betStatus'] == 'ACCEPTED' or bet['betStatus'] == 'CANCELLED' or bet['betStatus'] == 'WON' or bet['betStatus'] == 'REFUNDED' or bet['betStatus'] == 'PENDING_ACCEPTANCE'):
                return True
    
    return False


import pytest

from PS3838.utils.errors import RetrieveMatchError
from PS3838.utils.tools_code import verify_result

MATCH = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Argentina', 'away': 'Colombia', 'rotNum': '30805', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225}


def test_verify_result_team1_wins():
    assert verify_result(MATCH, "Argentina", "Colombia", 1) == 1

def test_verify_result_team2_wins():
    assert verify_result(MATCH, "Argentina", "Colombia", 2) == 2

def test_verify_result_draw():
    assert verify_result(MATCH, "Argentina", "Colombia", 0) == 0

def test_verify_result_team1_wins_with_result_2():
    assert verify_result(MATCH, "Colombia", "Argentina", 2) == 1

def test_verify_result_team2_wins_with_result_1():
    assert verify_result(MATCH, "Colombia", "Argentina", 1) == 2

def test_verify_result_no_result():
    assert verify_result(MATCH, "Argentina", "Colombia", None) is None

def test_verify_result_invalid_result():
    with pytest.raises(RetrieveMatchError):
        verify_result(MATCH, "Argentina", "Colombia", 3)

def test_verify_result_case_insensitive():
    assert verify_result(MATCH, "argentina", "colombia", 1) == 1

def test_verify_result_with_spaces():
    assert verify_result(MATCH, "Argen tina", "Col ombia", 1) == 1

def test_verify_result_with_punctuation():
    assert verify_result(MATCH, "Argentina!", "Colombia!", 1) == 1

def test_verify_result_partial_name():
    assert verify_result(MATCH, "Arg", "Colom", 1) == 1

def test_verify_result_mismatched_team_names():
    match =  {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Brazil', 'away': 'Chile', 'rotNum': '30805', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225}
    with pytest.raises(RetrieveMatchError):
        verify_result(match, "Argentina", "Colombia", 1)


from datetime import datetime
from unittest.mock import patch

import pytest

from PS3838 import betting
from PS3838.utils.constants import ID_SOCCER
from PS3838.utils.errors import CredentialError, ParameterError

CREDENTIALS = {"username": "test", "password": "test"}
LIST_MATCHES = [
    {'league': 'CONMEBOL - Copa America', 'team1': 'Colom', 'team2': 'Argentina', 'date': datetime(2024, 7, 15, 0, 0), 'result': 1, 'amount': 5, 'odd_min': 1.05},
    {'league': 'CONMEBOL - Copa America', 'team1': 'Canada', 'team2': 'Uruguay', 'date': datetime(2024, 7, 14, 0, 0), 'result': 2, 'amount': 10, 'odd_min': 1.10}
]
RETRIEVING_RESPONSE = [
    (
        {'id': 1593707406, 'league': 1872, 'result': 1, 'amount': 5, 'odd_min': 1.05, 'home': 'Argentina', 'away': 'Colombia', 'starts': '2024-07-15T00:00:00Z', 'line_id': 1},
        {'team1_odds': 2.17, 'draw_odds': 2.94, 'team2_odds': 4.41}
    ),
    (
        {'id': 1593707403, 'league': 1872, 'result': 2, 'amount': 10, 'odd_min': 1.10, 'home': 'Canada', 'away': 'Uruguay', 'starts': '2024-07-14T00:00:00Z', 'line_id': 2},
        {'team1_odds': 2.50, 'draw_odds': 3.20, 'team2_odds': 2.70}
    )
]

@patch('PS3838.PS3838.check_list_matches')
@patch('PS3838.PS3838.check_credentials')
@patch('PS3838.PS3838.retrieving')
@patch('PS3838.PS3838.Bet', autospec=True)
def test_betting_maintenance(mock_bet, mock_retrieving, mock_check_credentials, mock_check_list_matches):
    mock_bet_instance = mock_bet.return_value
    mock_retrieving.return_value = []
    
    mock_bet_instance.check_maintenance.return_value = {"status": "ALL_BETTING_CLOSED"}
    
    credentials = CREDENTIALS
    list_matches = LIST_MATCHES

    assert None == betting(credentials, list_matches)
    mock_check_credentials.assert_called_once_with(credentials)
    mock_check_list_matches.assert_called_once_with(list_matches, to_bet=True)
    mock_bet_instance.check_maintenance.assert_called_once()
    mock_bet_instance.place_bet.assert_not_called()
    mock_retrieving.assert_not_called()
    

@patch('PS3838.PS3838.check_credentials')
def test_betting_invalid_credentials(mock_check_credentials):
    mock_check_credentials.side_effect = CredentialError("Problem with the credentials")

    credentials = {"username": ""}
    list_matches = LIST_MATCHES
    
    with pytest.raises(CredentialError):
        betting(credentials, list_matches)

@patch('PS3838.utils.tools_check.check_list_matches')
def test_betting_invalid_matches(mock_check_list_matches):
    mock_check_list_matches.side_effect = ParameterError("Problem with the list of matches")

    credentials = CREDENTIALS
    list_matches = []
    
    with pytest.raises(ParameterError):
        betting(credentials, list_matches)


@patch('PS3838.PS3838.verify_odds')
@patch('PS3838.PS3838.is_already_bet')
@patch('PS3838.PS3838.check_list_matches')
@patch('PS3838.PS3838.check_credentials')
@patch('PS3838.PS3838.Retrieve', autospec=True)
@patch('PS3838.PS3838.Bet', autospec=True)
@patch('PS3838.PS3838.retrieving')
def test_betting_success(mock_retrieving, mock_bet, mock_retrieve, mock_check_credentials, mock_check_list_matches, mock_is_already_bet, mock_verify_odds):
    mock_bet_instance = mock_bet.return_value
    mock_retrieve_instance = mock_retrieve.return_value

    # Mock the check_maintenance method and the place_bet method
    mock_is_already_bet.return_value = False
    mock_verify_odds.return_value = True
    mock_bet_instance.check_maintenance.return_value = {"status": "ALL_BETTING_ENABLED"}
    mock_bet_instance.place_bet.return_value = {
    "status": "ACCEPTED",
    "errorCode": "ALL_BETTING_CLOSED",
    "betId": 0,
    "uniqueRequestId": "A9EB2EB1-13A5-4600-9F1B-4859379CDEC4",
    "betterLineWasAccepted": True,
    "price": 0
}

    # Mock the retrieving function
    mock_retrieving.return_value = RETRIEVING_RESPONSE
    
    credentials = CREDENTIALS
    list_matches = LIST_MATCHES
    
    betting(credentials, list_matches)
    
    mock_check_credentials.assert_called_once_with(credentials)
    mock_check_list_matches.assert_called_once_with(list_matches, to_bet=True)
    mock_bet.assert_called_once_with(credentials=credentials)
    mock_retrieve.assert_called_once_with(credentials=credentials)
    mock_bet_instance.check_maintenance.assert_called_once()
    mock_retrieving.assert_called_once_with(credentials, list_matches, mock_retrieve_instance)
    
    mock_bet_instance.place_bet.assert_any_call(
        odds_format="Decimal", stake=5, line_id=1, sport_id=ID_SOCCER, event_id=1593707406, period_number=0, bet_type="MONEYLINE", team="TEAM1"
    )
    mock_bet_instance.place_bet.assert_any_call(
        odds_format="Decimal", stake=10, line_id=2, sport_id=ID_SOCCER, event_id=1593707403, period_number=0, bet_type="MONEYLINE", team="TEAM2"
    )

@patch('PS3838.PS3838.verify_odds')
@patch('PS3838.PS3838.is_already_bet')
@patch('PS3838.PS3838.check_list_matches')
@patch('PS3838.PS3838.check_credentials')
@patch('PS3838.PS3838.Retrieve', autospec=True)
@patch('PS3838.PS3838.Bet', autospec=True)
@patch('PS3838.PS3838.retrieving')
def test_betting_exception_handling(mock_retrieving, mock_bet, mock_retrieve, mock_check_credentials, mock_check_list_matches, mock_is_already_bet, mock_verify_odds):
    mock_bet_instance = mock_bet.return_value
    mock_retrieve_instance = mock_retrieve.return_value
    mock_check_credentials.return_value = None
    mock_check_list_matches.return_value = None

    # Mock the check_maintenance method
    mock_bet_instance.check_maintenance.return_value = {"status": "ALL_BETTING_ENABLED"}
    mock_is_already_bet.return_value = False
    mock_verify_odds.return_value = True

    # Mock the retrieving function
    mock_retrieving.return_value = [
        (
            {'id': 1593707406, 'league': 1872, 'result': 1, 'amount': 5, 'odd_min': 1.05, 'home': 'Argentina', 'away': 'Colombia', 'starts': '2024-07-15T00:00:00Z', 'line_id': 1},
            {'team1_odds': 2.17, 'draw_odds': 2.94, 'team2_odds': 4.41}
        )
    ]
    
    # Mock the place_bet method to raise an exception
    mock_bet_instance.place_bet.side_effect = Exception("Test exception")

    credentials = {"username": "test", "password": "test"}
    list_matches = [
        {'id': 1, 'league': 'CONMEBOL - Copa America', 'team1': 'Colom', 'team2': 'Argentina', 'date': datetime(2024, 7, 15, 0, 0), 'result': 1, 'amount': 5, 'odd_min': 1.05}
    ]
    
    betting(credentials, list_matches, logger_active=True)
    
    mock_check_credentials.assert_called_once_with(credentials)
    mock_check_list_matches.assert_called_once_with(list_matches, to_bet=True)
    mock_bet.assert_called_once_with(credentials=credentials)
    mock_retrieve.assert_called_once_with(credentials=credentials)
    mock_bet_instance.check_maintenance.assert_called_once()
    mock_retrieving.assert_called_once_with(credentials, list_matches, mock_retrieve_instance)
    
    mock_bet_instance.place_bet.assert_called_once_with(
        odds_format="Decimal", stake=5, line_id=1, sport_id=ID_SOCCER, event_id=1593707406, period_number=0, bet_type="MONEYLINE", team="TEAM1"
    )
    # Ensure the exception is handled, and further calls are not made

@patch('PS3838.PS3838.is_already_bet')
@patch('PS3838.PS3838.check_list_matches')
@patch('PS3838.PS3838.check_credentials')
@patch('PS3838.PS3838.Retrieve', autospec=True)
@patch('PS3838.PS3838.Bet', autospec=True)
@patch('PS3838.PS3838.retrieving')
def test_betting_with_already_bet(mock_retrieving, mock_bet, mock_retrieve, mock_check_credentials, mock_check_list_matches, mock_is_already_bet):
    mock_bet_instance = mock_bet.return_value
    mock_retrieve_instance = mock_retrieve.return_value

    # Mock the check_maintenance method and the place_bet method
    mock_is_already_bet.return_value = True
    mock_bet_instance.check_maintenance.return_value = {"status": "ALL_BETTING_ENABLED"}

    # Mock the retrieving function
    mock_retrieving.return_value = RETRIEVING_RESPONSE
    
    credentials = CREDENTIALS
    list_matches = LIST_MATCHES
    
    betting(credentials, list_matches)
    
    mock_check_credentials.assert_called_once_with(credentials)
    mock_check_list_matches.assert_called_once_with(list_matches, to_bet=True)
    mock_bet.assert_called_once_with(credentials=credentials)
    mock_retrieve.assert_called_once_with(credentials=credentials)
    mock_bet_instance.check_maintenance.assert_called_once()
    mock_retrieving.assert_called_once_with(credentials, list_matches, mock_retrieve_instance)
    
    mock_bet_instance.place_bet.assert_not_called()

@patch('PS3838.PS3838.verify_odds')
@patch('PS3838.PS3838.is_already_bet')
@patch('PS3838.PS3838.check_list_matches')
@patch('PS3838.PS3838.check_credentials')
@patch('PS3838.PS3838.Retrieve', autospec=True)
@patch('PS3838.PS3838.Bet', autospec=True)
@patch('PS3838.PS3838.retrieving')
def test_betting_without_verify_odds(mock_retrieving, mock_bet, mock_retrieve, mock_check_credentials, mock_check_list_matches, mock_is_already_bet, mock_verify_odds):
    mock_bet_instance = mock_bet.return_value
    mock_retrieve_instance = mock_retrieve.return_value
    mock_check_list_matches.return_value = None
    mock_check_credentials.return_value = None

    # Mock the check_maintenance method and the place_bet method
    mock_is_already_bet.return_value = False
    mock_verify_odds.return_value = False
    mock_bet_instance.check_maintenance.return_value = {"status": "ALL_BETTING_ENABLED"}

    # Mock the retrieving function
    mock_retrieving.return_value = RETRIEVING_RESPONSE
    
    credentials = CREDENTIALS
    list_matches = LIST_MATCHES
    
    betting(credentials, list_matches)
    
    mock_check_credentials.assert_called_once_with(credentials)
    mock_check_list_matches.assert_called_once_with(list_matches, to_bet=True)
    mock_bet.assert_called_once_with(credentials=credentials)
    mock_retrieve.assert_called_once_with(credentials=credentials)
    mock_bet_instance.check_maintenance.assert_called_once()
    mock_retrieving.assert_called_once_with(credentials, list_matches, mock_retrieve_instance)
    
    mock_bet_instance.place_bet.assert_not_called()


from datetime import datetime

import pytest

from PS3838.utils.errors import RetrieveMatchError
from PS3838.utils.tools_code import find_match

FIXTURES = [{'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Argentina', 'away': 'Colombia', 'rotNum': '30805', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225}, {'id': 1593707403, 'starts': '2024-07-14T00:00:00Z', 'home': 'Canada', 'away': 'Uruguay', 'rotNum': '30801', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538881047}]


def test_find_match_exact_match():
    match = find_match(FIXTURES, "Argentina", "Colombia")
    assert match["id"] == 1593707406

def test_find_match_with_date():
    match_date = datetime(2024, 7, 15)
    match = find_match(FIXTURES, "Argentina", "Colombia", match_date)
    assert match["id"] == 1593707406

def test_find_match_case_insensitive():
    match = find_match(FIXTURES, "argentina", "colombia")
    assert match["id"] == 1593707406

def test_find_match_no_match():
    with pytest.raises(RetrieveMatchError):
        find_match(FIXTURES, "Chile", "Uruguay")

def test_find_match_with_similar_names():
    match = find_match(FIXTURES, "Argentin", "Colombi")
    assert match["id"] == 1593707406
    
    


from datetime import datetime
from unittest.mock import Mock

import pytest

from PS3838.PS3838Retrieve import Retrieve
from PS3838.utils.errors import RetrieveMatchError
from PS3838.utils.tools_code import retrieve_match

LEAGUE_RESPONSE = {
    "leagues": [
        {"id": 1872, "name": "CONMEBOL - Copa America"},
        {"id": 1873, "name": "La Liga"}
    ]
}

FIXTURES_RESPONSE = {
    "league": [
        {
            "events": [
                {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Argentina', 'away': 'Colombia', 'rotNum': '30805', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225}, {'id': 1593707403, 'starts': '2024-07-14T00:00:00Z', 'home': 'Canada', 'away': 'Uruguay', 'rotNum': '30801', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538881047}
            ]
        }
    ]
}

FIXTURES_RESPONSE_2 = {
    "league": [
        {
            "events": [
                {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Argentina', 'away': 'Colombia', 'rotNum': '30805', 'liveStatus': 2, 'status': 'C', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225}, {'id': 1593707403, 'starts': '2024-07-14T00:00:00Z', 'home': 'Canada', 'away': 'Uruguay', 'rotNum': '30801', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538881047}
            ]
        }
    ]
}


def test_retrieve_match_success():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Colom',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    match = retrieve_match(
        mock_api_retrieve,
        entry['league'],
        entry['team1'],
        entry['team2'],
        entry['date'],
        entry['result'],
        entry['amount'],
        entry['odd_min']
    )

    assert match["id"] == 1593707406
    assert match["league"] == 1872
    assert match["result"] == 1
    assert match["amount"] == 5
    assert match["odd_min"] == 1.05

def test_retrieve_match_no_match():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Venezuela',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )

def test_retrieve_match_without_date():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Colom',
        'team2': 'Argentina',
        'result': 1,
        'amount': 5,
        'odd_min': 1.05
    }

    match = retrieve_match(
        mock_api_retrieve,
        entry['league'],
        entry['team1'],
        entry['team2'],
        result = entry['result'],
        amount = entry['amount'],
        odd_min = entry['odd_min']
    )

    assert match["id"] == 1593707406
    assert match["league"] == 1872
    assert match["result"] == 2
    assert match["amount"] == 5
    assert match["odd_min"] == 1.05

def test_retrieve_match_non_existent_league():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE

    entry = {
        'id': 1,
        'league': 'Non Existent League',
        'team1': 'Colom',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )

def test_retrieve_match_non_existent_league():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE

    entry = {
        'id': 1,
        'league': 'Non Existent League',
        'team1': 'Colom',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )

def test_retrieve_match_teams_not_in_fixture():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Brazil',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )

def test_retrieve_match_empty_fixtures():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = {"league": [{"events": []}]}  # Empty fixtures

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Colom',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )


def test_retrieve_match_fixture_status_other_than_O():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE_2

    entry = {
        'id': 1,
        'league': 'CONMEBOL - Copa America',
        'team1': 'Colom',
        'team2': 'Argentina',
        'date': datetime(2024, 7, 15, 0, 0),
        'result': 2,
        'amount': 5,
        'odd_min': 1.05
    }

    with pytest.raises(RetrieveMatchError):
        retrieve_match(
            mock_api_retrieve,
            entry['league'],
            entry['team1'],
            entry['team2'],
            entry['date'],
            entry['result'],
            entry['amount'],
            entry['odd_min']
        )



import pytest

from PS3838.utils.tools_check import CredentialError, check_credentials


def test_check_credentials_success():
    credentials = {"username": "test_user", "password": "test_password"}
    try:
        check_credentials(credentials)
    except CredentialError:
        pytest.fail("check_credentials raised CredentialError unexpectedly!")

def test_check_credentials_missing_credentials():
    with pytest.raises(CredentialError, match="Credentials are missing"):
        check_credentials(None)

def test_check_credentials_not_a_dict():
    with pytest.raises(CredentialError, match="Credentials must be a dictionary"):
        check_credentials("username=test_user&password=test_password")

def test_check_credentials_missing_keys():
    credentials = {"username": "test_user"}
    with pytest.raises(CredentialError, match="Credentials must contain 'username' and 'password' keys"):
        check_credentials(credentials)

    credentials = {"password": "test_password"}
    with pytest.raises(CredentialError, match="Credentials must contain 'username' and 'password' keys"):
        check_credentials(credentials)

def test_check_credentials_keys_not_strings():
    credentials = {"username": 123, "password": "test_password"}
    with pytest.raises(CredentialError, match="Credentials 'username' and 'password' must be strings"):
        check_credentials(credentials)

    credentials = {"username": "test_user", "password": 456}
    with pytest.raises(CredentialError, match="Credentials 'username' and 'password' must be strings"):
        check_credentials(credentials)



from datetime import datetime
from unittest.mock import Mock

from PS3838.PS3838Retrieve import Retrieve
from PS3838.utils.tools_code import retrieve_matches

LEAGUE_RESPONSE = {
    "leagues": [
        {"id": 1872, "name": "CONMEBOL - Copa America"},
        {"id": 1873, "name": "La Liga"}
    ]
}

FIXTURES_RESPONSE = {
    "league": [
        {
            "events": [
                {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z', 'home': 'Argentina', 'away': 'Colombia', 'rotNum': '30805', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538876225},
                {'id': 1593707403, 'starts': '2024-07-14T00:00:00Z', 'home': 'Canada', 'away': 'Uruguay', 'rotNum': '30801', 'liveStatus': 2, 'status': 'O', 'parlayRestriction': 2, 'altTeaser': False, 'resultingUnit': 'Regular', 'betAcceptanceType': 0, 'version': 538881047}
            ]
        }
    ]
}


def test_retrieve_matches_success():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    list_matches = [
        {
            'id': 1,
            'league': 'CONMEBOL - Copa America',
            'team1': 'Colom',
            'team2': 'Argentina',
            'date': datetime(2024, 7, 15, 0, 0),
            'result': 2,
            'amount': 5,
            'odd_min': 1.05
        },
        {
            'id': 2,
            'league': 'CONMEBOL - Copa America',
            'team1': 'Canada',
            'team2': 'Uruguay',
            'date': datetime(2024, 7, 14, 0, 0),
            'result': 1,
            'amount': 10,
            'odd_min': 1.10
        }
    ]

    matches = retrieve_matches(list_matches, mock_api_retrieve)

    assert len(matches) == 2
    assert matches[0]["id"] == 1593707406
    assert matches[0]["league"] == 1872
    assert matches[0]["result"] == 1
    assert matches[0]["amount"] == 5
    assert matches[0]["odd_min"] == 1.05
    assert matches[1]["id"] == 1593707403
    assert matches[1]["league"] == 1872
    assert matches[1]["result"] == 1
    assert matches[1]["amount"] == 10
    assert matches[1]["odd_min"] == 1.10

def test_retrieve_matches_partial_success():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    list_matches = [
        {
            'id': 1,
            'league': 'CONMEBOL - Copa America',
            'team1': 'Colom',
            'team2': 'Argentina',
            'date': datetime(2024, 7, 15, 0, 0),
            'result': 2,
            'amount': 5,
            'odd_min': 1.05
        },
        {
            'id': 2,
            'league': 'CONMEBOL - Copa America',
            'team1': 'NonExistent',
            'team2': 'Uruguay',
            'date': datetime(2024, 7, 14, 0, 0),
            'result': 1,
            'amount': 10,
            'odd_min': 1.10
        }
    ]

    matches = retrieve_matches(list_matches, mock_api_retrieve)

    assert len(matches) == 1
    assert matches[0]["id"] == 1593707406
    assert matches[0]["league"] == 1872
    assert matches[0]["result"] == 1
    assert matches[0]["amount"] == 5
    assert matches[0]["odd_min"] == 1.05


def test_retrieve_matches_empty_list():
    mock_api_retrieve = Mock(spec=Retrieve)
    list_matches = []

    matches = retrieve_matches(list_matches, mock_api_retrieve)

    assert matches == []

def test_retrieve_matches_no_date():
    mock_api_retrieve = Mock(spec=Retrieve)
    mock_api_retrieve.get_leagues_v3.return_value = LEAGUE_RESPONSE
    mock_api_retrieve.get_fixtures_v3.return_value = FIXTURES_RESPONSE

    list_matches = [
        {
            'id': 1,
            'league': 'CONMEBOL - Copa America',
            'team1': 'Colom',
            'team2': 'Argentina',
            'result': 2,
            'amount': 5,
            'odd_min': 1.05
        },
        {
            'id': 2,
            'league': 'CONMEBOL - Copa America',
            'team1': 'Canada',
            'team2': 'Uruguay',
            'result': 1,
            'amount': 10,
            'odd_min': 1.10
        }
    ]

    matches = retrieve_matches(list_matches, mock_api_retrieve)

    assert len(matches) == 2
    assert matches[0]["id"] == 1593707406
    assert matches[0]["league"] == 1872
    assert matches[0]["result"] == 1
    assert matches[0]["amount"] == 5
    assert matches[0]["odd_min"] == 1.05
    assert matches[1]["id"] == 1593707403
    assert matches[1]["league"] == 1872
    assert matches[1]["result"] == 1
    assert matches[1]["amount"] == 10
    assert matches[1]["odd_min"] == 1.10


from unittest.mock import patch

from PS3838.utils.tools_code import is_already_bet


@patch('PS3838.utils.tools_code.Bet')
def test_is_already_bet_true(mock_api_bet):
    match = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z'}
    
    mock_api_bet.get_bets_v3.return_value = {
        "straightBets": [
            {"eventId": 1593707406, "betStatus": "ACCEPTED"}
        ]
    }
    
    assert is_already_bet(mock_api_bet, match) == True

@patch('PS3838.utils.tools_code.Bet')
def test_is_already_bet_false_no_bets(mock_api_bet):
    match = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z'}
    
    mock_api_bet.get_bets_v3.return_value = {}
    
    assert is_already_bet(mock_api_bet, match) == False
    
@patch('PS3838.utils.tools_code.Bet')
def test_is_already_bet_false_different_event_id(mock_api_bet):
    match = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z'}
    
    mock_api_bet.get_bets_v3.return_value = {
        "straightBets": [
            {"eventId": 1234567890, "betStatus": "ACCEPTED"}
        ]
    }
    
    assert is_already_bet(mock_api_bet, match) == False

@patch('PS3838.utils.tools_code.Bet')
def test_is_already_bet_false_different_bet_status(mock_api_bet):
    match = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z'}
    
    mock_api_bet.get_bets_v3.return_value = {
        "straightBets": [
            {"eventId": 1593707406, "betStatus": "PENDING"}
        ]
    }
    
    assert is_already_bet(mock_api_bet, match) == False

@patch('PS3838.utils.tools_code.Bet')
def test_is_already_bet_no_straight_bets_key(mock_api_bet):
    match = {'id': 1593707406, 'starts': '2024-07-15T00:00:00Z'}
    
    mock_api_bet.get_bets_v3.return_value = {}
    
    assert is_already_bet(mock_api_bet, match) == False
